/* gameprob.c - stochastic model for the game results
 *
 * Copyright (C) 2002, 2003  Jochen Voss <voss@mathematik.uni-kl.de>
 *
 * This module implements a stochastic model for the outcome
 * of a game of go.  The functionality is in the function
 * 'probability' below.
 *
 * These functions are hooked into the Python scripts via a
 * SWIG-generated interface.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

static const  char  rcsid[] = "$Id: gameprob.c 5068 2003-07-19 12:43:00Z voss $";

#include <stdlib.h>
#include <string.h>
#include <search.h>
#include <setjmp.h>
#include <math.h>
#include <assert.h>

#include "gameprob.h"


/* Jigo probabilities for games with integral komi values.
 * These are only wild guesses, which are not based on facts.  */
#define P_JIGO_19 0.008
#define P_JIGO_13 0.017
#define P_JIGO_9 0.035

/* variable internal parameters */
static  double /*1,*/ h13, h9;	/* h = grade difference / handicap stone */
static  double  k19, k13, k9;	/* k = komi / handicap stone */
static  double  e19, e13, e9;	/* komi appropriate for even game */
static  double  jo19, jo13, jo9; /* jigo offsets */
static  double  s19, s13, s9;	/* variance of results */

static const struct param_info internal [] = {
  { -1, "e19", &e19, 6, 0, 12, 19 },
  { -2, "e13", &e13, 6, 0, 12, 13 },
  { -3, "e9", &e9, 4, 0, 12, 9 },
  { -4, "jo19", &jo19, 0.01, 0, 0.1, 19 },
  { -5, "jo13", &jo13, 0.02, 0, 0.1, 13 },
  { -6, "jo9", &jo9, 0.04, 0, 0.1, 9 },
  { -7, "h13", &h13, 2.5, 1, 10, 13 },
  { -8, "h9", &h9, 5, 1, 10, 9 },
  { -9, "k19", &k19, 10, 1, 20, 19 },
  { -10, "k13", &k13, 10, 5, 30, 13 },
  { -11, "k9", &k9, 13, 5, 20, 9 },
  { -12, "s19", &s19, 5, 2, 15, 19 },
  { -13, "s13", &s13, 5, 2, 15, 13 },
  { -14, "s9", &s9, 5, 2, 15, 9 },
};
#define  PARAM_COUNT  (sizeof(internal)/sizeof(struct param_info))

/**********************************************************************
 * game model parameters.
 */

static void *parameters = NULL;

static int
param_comp (const void *a, const void *b)
{
  const struct param_info *pa = a;
  const struct param_info *pb = b;
  return  (pa->id > pb->id) - (pa->id < pb->id);
}

struct param_info *
find_parameter (int id)
{
  struct param_info pi, **ptr;;

  pi.id = id;
  ptr = tfind (&pi, &parameters, param_comp);
  return  ptr ? *ptr : NULL;
}

static double
Phi (double x)
/* The distribution function of the standard normal distribution.  */
{
  return  (1+erf (x/M_SQRT2))/2;
}

static double
jigo_offset (double p_jigo)
/* Return the value x with 'Phi(+x)-Phi(-x) == P_JIGO'.
 * P_JIGO is the jigo probability for even game.  */
{
  double  l, r, q;

  q = 0.5*(1+p_jigo);
  assert (0.5 <= q && q < Phi(1));
  l = 0;  r = 1;
  while (l+1e-6 < r) {
    double  m;
    assert (Phi(l) <= q && q < Phi(r));
    m = 0.5*(l+r);
    if (Phi(m) <= q) {
      l = m;
    } else {
      r = m;
    }
  }
  return  l;
}

void
init_parameters (void)
{
  int  i;

  jo9 = jigo_offset (P_JIGO_9);
  jo13 = jigo_offset (P_JIGO_13);
  jo19 = jigo_offset (P_JIGO_19);
  
  for (i=0; i<PARAM_COUNT; ++i) {
    *internal[i].ptr = internal[i].default_value;
    tsearch (internal+i, &parameters, param_comp);
  }
}

void
register_parameter (int id, const char *name)
{
  struct param_info *ptr;

  assert (id >= 0 && ! find_parameter (id));

  ptr = malloc (sizeof(struct param_info));
  ptr->id = id;
  ptr->name = name ? strdup(name) : NULL;
  ptr->ptr = malloc(sizeof(double));
  ptr->default_value = -19;
  ptr->min = -45;
  ptr->max = 7;
  ptr->board = 0;
  *(ptr->ptr) = ptr->default_value;
  tsearch (ptr, &parameters, param_comp);
}

static const char *needle;
struct param_info *name_search_res;
static jmp_buf trampoline;

static void
name_search_fn (const void *node, VISIT val, int level)
{
  if (val == leaf || val == preorder) {
    const struct param_info *p = *( struct param_info *const*)node;
    if (strcmp (p->name, needle) == 0) {
      name_search_res = (struct param_info *)p;
      longjmp (trampoline, 1);
    }
  }
}

struct param_info *
find_param_by_name (const char *name)
/* Note: this function is not thread safe.  */
{
  needle = name;
  if (setjmp (trampoline)) {
    return  name_search_res;
  } else {
    twalk (parameters, name_search_fn);
    return  NULL;
  }
}

void
set_param (int id, double val)
{
  struct param_info *p = find_parameter (id);
  if (val < p->min) {
    val = p->min;
  } else if (val > p->max) {
    val = p->max;
  }
  *(p->ptr) = val;
}

double
get_param (int id)
{
  struct param_info *p = find_parameter (id);
  return *(p->ptr);
}

/**********************************************************************
 * Games
 */

static void *games = NULL;

static int
game_comp (const void *a, const void *b)
{
  const struct game_info *pa = a;
  const struct game_info *pb = b;
  return  (pa->id > pb->id) - (pa->id < pb->id);
}

struct game_info *
find_game (int id)
{
  struct game_info pi, **ptr;;

  pi.id = id;
  ptr = tfind (&pi, &games, game_comp);
  return  ptr ? *ptr : NULL;
}

void
register_game (int id, int black, int white, int board,
	       int hc, int komi, int with_jigo,
	       const char *res)
{
  struct game_info *ptr;

  assert (! find_game(id));

  ptr = malloc (sizeof(struct game_info));
  ptr->id = id;
  ptr->black = black;
  ptr->white = white;
  ptr->board = board;
  ptr->hc = hc;
  ptr->komi = komi;
  ptr->with_jigo = with_jigo;
  if (strcmp(res,"black")==0) {
    ptr->res = res_BLACK;
  } else if (strcmp(res,"white")==0) {
    ptr->res = res_WHITE;
  } else if (strcmp(res,"jigo")==0) {
    ptr->res = res_JIGO;
  } else {
    assert (0);
  }
  tsearch (ptr, &games, game_comp);
}

static double
game_offset (int board, int handicap, int komi)
/* Returns the strength offset, measured in player grades.
 * Positive values indicate, that the setting favours black.  */
{
  double  h, k, e;

  switch (board) {
  case 19:
    h = 1;  k = k19;  e = e19;
    break;
  case 13:
    h = h13;  k = k13;  e = e13;
    break;
  case 9:
    h = h9;  k = k9;  e = e9;
    break;
  default:
    abort ();
  }

  if (handicap>0)  return h*(handicap-1 + (e-komi)/k);
  return  h*(e-komi)/k;
}

double
probability (int black, int white, int board,
	     int hc, int komi, int with_jigo, enum result res)
/* Return the probability of a game.
 * BLACK and WHITE are indices for the player strength values table.
 * The return value depends on the model parameters.  */
{
  double  mu, sigma, jo;
  double  p;

  switch (board) {
  case 19:
    sigma = s19; jo = jo19;
    break;
  case 13:
    sigma = s13; jo = jo13;
    break;
  case 9:
    sigma = s9; jo = jo9;
    break;
  default:
    assert (0);
  }

  mu = (get_param(black)-get_param(white)) + game_offset (board, hc, komi);

  switch (res) {
  case res_BLACK:
    p = 1 - Phi ((-mu+jo)/sigma);
    break;
  case res_WHITE:
    if (with_jigo) {
      p = Phi ((-mu-jo)/sigma);
    } else {
      p = Phi ((-mu+jo)/sigma);
    }
    break;
  default:			/* res_JIGO */
    assert (with_jigo);
    p = Phi ((-mu+jo)/sigma) - Phi ((-mu-jo)/sigma);
    break;
  }
  return  p;
}

double
game_prob (int game)
/* Return the probability for the outcome of GAME.
 * The game index refers to an 'id' argument value of the
 * 'register_game' function.  The result depends on the
 * model parameters.  */
{
  struct game_info *g = find_game (game);
  return  probability (g->black, g->white,
		       g->board, g->hc, g->komi, g->with_jigo, g->res);
}

int
applies_to (int param, int game)
/* Check whether a given model parameter affect the probability of GAME.
 * Positive parameter indices indicate external parameters (i.e. player
 * strengths), negative indices indicate internal model parameters from
 * the table 'internal'.  */
{
  struct game_info *g = find_game (game);
  if (param >= 0) {
    return  (g->black == param || g->white == param);
  } else {
    struct param_info *p = find_parameter (param);
    return  (p->board == 0 || p->board == g->board);
  }
}
